require 'ruby/proto/version'

module Ruby
  module Proto
    class << self
      def testForRequire()
        true
      end

      def toProtoData(_hash, _proc)
        protoData(_proc, _hash)
        _proc
      end

      private
      def getBasicField
        @arrBasicField ||= ["int32", "uint32", "string", "bool"]
      end

      def convert_basic_field_value(_field, _value)
        return _value if _value.is_a?(Array)
        return _value if _value.is_a?(Hash)
        index = getBasicField.index(_field.type.to_s)
        unless index.nil?
          _value = case index
                     when 0..1
                       _value.to_i
                     when 2
                       _value.to_s
                     when 3
                       _value == 1
                     else
                       _value
                   end
        end
        _value
      end

      def protoData(_proc, _hash)
        _proc.fields.each do |_tag, _field|
          ret = _hash.fetch(_field.name.to_s, nil)
          next if ret.nil?
          if ret.is_a?(Hash)
            subProc = _field.type.new
            protoData(subProc, ret)
            ret = subProc
          elsif ret.is_a?(Array)
            arr = []
            ret.each do |_hash|
              subProc = nil
              if (getBasicField.include?(_field.type.to_s))
                subProc = convert_basic_field_value(_field, _hash)
              else
                subProc = _field.type.new
                protoData(subProc, _hash)
              end
              arr.push subProc
            end
            ret = arr
          end
          _proc.send(_field.name.to_s + "=", ret)
        end
      end
    end
  end

  module SerialData
    class << self
      def intToByteArray(_value, _length = 4)
        arr = []
        _length.times do |_index|
          val = _value & 0xff;
          _value = _value >> 8;
          arr.push(val)
        end
        arr
      end

      def byteArrayToInt(_arr)
        value = 0
        _arr.reverse_each do |_byte|
          # puts "#{__callee__} byte:[#{_byte}]"
          value = value << 8
          value = value | _byte
        end
        value
      end
    end

    def storeInt32(_value)
      doStoreInt_(_value, 4)
    end

    def storeInt16(_value)
      doStoreInt_(_value, 2)
    end

    def storeInt8(_value)
      doStoreInt_(_value, 1)
    end

    def loadInt8
      doLoadInt_(1)
    end

    def loadInt16
      doLoadInt_(2)
    end

    def loadInt32
      doLoadInt_(4)
    end

    private
    def doStoreInt_(_value, _length)
      arr = SerialData.intToByteArray(_value, _length)
      str = arr.pack("C#{_length}")
      writeDataImp(str)
    end

    def doLoadInt_(_length)
      str = readDataImp(_length)
      return 0 if str.nil? || str.empty?
      _length = str.length
      arr = str.unpack("C#{_length}")
      SerialData.byteArrayToInt(arr)
    end
  end

  class SerialString
    include SerialData
    attr_reader :str, :read_index

    def initialize(_value = "")
      resetString(_value)
    end

    def resetString(_value = "")
      @str = ""
      @str.encode!("binary")
      _value = _value || ""
      @str <<_value
      @read_index = 0
    end

    private
    def writeDataImp(_str)
      @str << _str
    end

    def readDataImp(_length)
      str = @str.slice(@read_index, _length)
      return str if str.nil? || str.empty?
      _length = str.length
      @read_index += _length
      str
    end
  end

end
